# -*- coding: utf-8 -*-

"""

dryjq.__main__

Module execution entry point

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import sys

from dryjq import cli

sys.exit(cli.main())

# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

dryjq.access

Data structure access using Path objects

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import copy
import logging

from typing import Any


class PathComponent:

    """single index component of a Path"""

    def __init__(self, index: Any, in_subscript: bool = False) -> None:
        """Initialize with the index"""
        self.__index = index
        self.__allow_lists = in_subscript
        if not in_subscript:
            self.__reason = "not in subscript"
        if not isinstance(index, int):
            self.__allow_lists = False
            self.__reason = "not a valid list index"
        #

    def get_value(self, data_structure: Any) -> Any:
        """Get the value from data_structure[self.__index]"""
        if not isinstance(data_structure, (list, dict)):
            raise TypeError(
                f"Index {self.__index!r} is not suitable for scalar value"
                f" {data_structure!r}!"
            )
        #
        if isinstance(data_structure, list) and not self.__allow_lists:
            raise TypeError(
                f"Index {self.__index!r} is not suitable for lists"
                f" ({self.__reason})!"
            )
        #
        try:
            value = data_structure[self.__index]
        except (IndexError, KeyError) as error:
            raise ValueError(
                f"Index {self.__index!r} not found in {data_structure!r}!"
            ) from error
        #
        logging.debug("Value at %r is %r", self.__index, value)
        return value

    def replace_value(self, data_structure: Any, new_value: Any) -> None:
        """Modify data_structure in place:
        replace the value at the existing index by new_value
        """
        # Ensure the index aready exists
        self.get_value(data_structure)
        data_structure[self.__index] = new_value

    def __repr__(self) -> str:
        """Return a representation"""
        if self.__allow_lists:
            purpose = "for maps and lists"
        else:
            purpose = "for maps only"
        #
        return f"<Index {self.__index!r} ({purpose})>"


class Path:

    """Address in data structures,
    a sequence of PathComponent instances.
    """

    def __init__(self, *components: PathComponent) -> None:
        """Initialize the internal components"""
        self.__complete_path = components

    def get_value(self, data_structure: Any) -> Any:
        """Get a value or substructure from data_structure
        using self.__complete_path
        """
        current_value = data_structure
        for element in self.__complete_path:
            current_value = element.get_value(current_value)
        #
        return current_value

    def replace_value(self, data_structure: Any, new_value: Any) -> Any:
        """Replace a value in the data structure,
        and return a copy with the replaced value
        """
        if not self.__complete_path:
            return new_value
        #
        ds_copy = copy.deepcopy(data_structure)
        current_value = ds_copy
        for element in self.__complete_path[:-1]:
            current_value = element.get_value(current_value)
        #
        last_element = self.__complete_path[-1]
        last_element.replace_value(current_value, new_value)
        return ds_copy


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

dryjq.cli

Command line interface

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import argparse
import logging

import sys

from typing import Any, IO, List, Optional, Type

import yaml

import dryjq

from dryjq import access
from dryjq import queries
from dryjq import streams


#
# Constants
#


RETURNCODE_OK = 0
RETURNCODE_ERROR = 1


#
# Functions
#


def execute_query(
    handler_class: Type[streams.StreamHandler],
    stream: IO,
    data_path: access.Path,
    replacement_value: Optional[str] = None,
    **output_options: Any,
) -> int:
    """Execute the query, write output and return the returncode"""
    try:
        file_handler = handler_class(stream)
    except (yaml.YAMLError, yaml.composer.ComposerError) as error:
        for line in str(error).splitlines():
            logging.error(line)
        #
        return RETURNCODE_ERROR
    #
    try:
        file_handler.execute_single_query(
            data_path, replacement_value=replacement_value
        )
        file_handler.write_output(**output_options)
    except (TypeError, ValueError, yaml.YAMLError) as error:
        for line in str(error).splitlines():
            logging.error(line)
        #
        return RETURNCODE_ERROR
    #
    return RETURNCODE_OK


def main(args: Optional[List[str]] = None) -> int:
    """Parse command line arguments and execute the matching function"""
    main_parser = argparse.ArgumentParser(
        prog="dryjq",
        description="Drastically Reduced YAML / JSON Query",
    )
    main_parser.set_defaults(
        loglevel=logging.INFO, query=".", output_indent=dryjq.DEFAULT_INDENT
    )
    main_parser.add_argument(
        "-v",
        "--verbose",
        action="store_const",
        const=logging.DEBUG,
        dest="loglevel",
        help="Output all messages including debug level",
    )
    main_parser.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        const=logging.ERROR,
        dest="loglevel",
        help="Limit message output to errors",
    )
    main_parser.add_argument(
        "--version",
        action="store_true",
        help="Print version and exit",
    )
    main_parser.add_argument(
        "--modify-in-place",
        action="store_true",
        help="Modify the input file in place instead of writing"
        " the result to standard output.",
    )
    output_group = main_parser.add_argument_group(
        "Output options", "control how output will be formatted"
    )
    output_group.add_argument(
        "-of",
        "--output-format",
        type=str.upper,
        choices=dryjq.SUPPORTED_FORMATS,
        help="File format. By default, the detected input format is used.",
    )
    output_group.add_argument(
        "-oi",
        "--output-indent",
        choices=(2, 4, 8),
        type=int,
        help="Indentation depth of blocks, in spaces (default: %(default)s).",
    )
    output_group.add_argument(
        "-osk",
        "--output-sort-keys",
        action="store_true",
        help="Sort mapping keys."
        " By default, mapping keys are left in input order.",
    )
    main_parser.add_argument(
        "query",
        nargs="?",
        help="The query (simplest form of yq/jq syntax,"
        " default is %(default)r).",
    )
    main_parser.add_argument(
        "input_file",
        nargs="?",
        help="The input file name."
        " By default, data will be read from standard input.",
    )
    arguments = main_parser.parse_args(args)
    if arguments.version:
        print(dryjq.__version__)
        return RETURNCODE_OK
    #
    logging.basicConfig(
        format="%(levelname)-8s | (%(funcName)s:%(lineno)s) %(message)s",
        level=arguments.loglevel,
    )
    data_path, replacement_value = queries.Parser().parse_query(
        arguments.query
    )
    output_options = dict(
        output_format=arguments.output_format,
        indent=arguments.output_indent,
        sort_keys=arguments.output_sort_keys,
    )
    if arguments.input_file is None:
        if arguments.modify_in_place:
            logging.warning("Cannot modify <stdin> in place")
        #
        return execute_query(
            streams.StreamHandler,
            sys.stdin,
            data_path,
            replacement_value=replacement_value,
            **output_options,
        )
    #
    file_helper = streams.FileHelper(
        arguments.input_file,
        modify_in_place=arguments.modify_in_place,
        replace_value_mode=(replacement_value is not None),
    )
    with file_helper.open(encoding="utf-8") as input_file:
        file_helper.lock(input_file)
        returncode = execute_query(
            file_helper.handler_class,
            input_file,
            data_path,
            replacement_value=replacement_value,
            **output_options,
        )
        file_helper.unlock(input_file)
    #
    return returncode


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

dryjq.streams

Stream and file handlers

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import json
import logging
import sys

try:
    import fcntl
except ModuleNotFoundError:
    pass
#

from typing import Any, IO, Optional, Type

import yaml

import dryjq

from dryjq import access


class StreamHandler:

    """YAML or JSON stream reader and data handler"""

    def __init__(self, stream_io: IO) -> None:
        """Read all data from the stream"""
        contents = stream_io.read()
        try:
            self.__filtered_data = json.loads(contents)
        except json.JSONDecodeError:
            self.__filtered_data = yaml.safe_load(contents)
            self.__input_format = dryjq.FORMAT_YAML
        else:
            self.__input_format = dryjq.FORMAT_JSON
        #
        self.__original_contents = contents
        self.__stream_io = stream_io

    @property
    def stream_io(self) -> IO:
        """Return the stream handle"""
        return self.__stream_io

    @property
    def original_contents(self) -> str:
        """Return the original contents"""
        return self.__original_contents

    def dump_data(
        self,
        output_format: Optional[str] = None,
        indent: int = dryjq.DEFAULT_INDENT,
        sort_keys: bool = False,
    ) -> str:
        """Return a dump of self.__filtered_data"""
        if output_format is None:
            output_format = self.__input_format
        #
        if output_format == dryjq.FORMAT_JSON:
            return json.dumps(
                self.__filtered_data,
                indent=indent,
                ensure_ascii=True,
                sort_keys=sort_keys,
            )
        #
        output = yaml.safe_dump(
            self.__filtered_data,
            allow_unicode=True,
            default_flow_style=False,
            indent=indent,
            sort_keys=sort_keys,
            explicit_end=False,
        )
        if isinstance(self.__filtered_data, (dict, list)):
            return output
        #
        if output.rstrip().endswith("\n..."):
            output = output.rstrip()[:-3]
        #
        return output

    def execute_single_query(
        self,
        data_path: access.Path,
        replacement_value: Optional[str] = None,
    ) -> None:
        """Execute the provided query to filter
        self.__filtered_data
        """
        logging.debug("Filtered data before executing the query:")
        logging.debug("%r", self.__filtered_data)
        if replacement_value is None:
            self.__filtered_data = data_path.get_value(self.__filtered_data)
        else:
            self.__filtered_data = data_path.replace_value(
                self.__filtered_data, yaml.safe_load(replacement_value.strip())
            )
        #
        logging.debug("Filtered data after executing the query:")
        logging.debug("%r", self.__filtered_data)

    def write_output(self, **output_options: Any) -> None:
        """Write output to stdout"""
        data_dump = self.dump_data(**output_options).rstrip()
        sys.stdout.write(f"{data_dump}\n")


class FileReader(StreamHandler):

    """File reader"""

    def __init__(self, stream_io: IO) -> None:
        """Go to the first byte in the file,
        then act as a normal StreamReader instance
        """
        stream_io.seek(0)
        super().__init__(stream_io)


class FileWriter(FileReader):

    """File writer
    Required locking can be done using the FileHelper class.
    """

    def write_output(self, **output_options: Any) -> None:
        """Write output to the file if data has changed"""
        data_dump = self.dump_data(**output_options)
        if data_dump == self.original_contents:
            logging.warning("File contents did not change.")
            return
        #
        self.stream_io.seek(0)
        self.stream_io.truncate(0)
        self.stream_io.write(data_dump)


class FileHelper:

    """File handling helper class"""

    no_lock_operation = 0

    def __init__(
        self,
        file_name: Optional[str],
        modify_in_place: bool = False,
        replace_value_mode: bool = False,
    ) -> None:
        """Determine whether the file should be opened in r
        in r or r+ mode, then return a tuple containing
        the file handler class, the open mode for the file
        and the lock and unlock operation magic numbers
        """
        if file_name is None:
            raise ValueError("No file name supplied!")
        #
        self.file_name = file_name
        exclusive_lock = self.no_lock_operation
        shared_lock = self.no_lock_operation
        unlock_operation = self.no_lock_operation
        try:
            exclusive_lock = fcntl.LOCK_EX
        except NameError:
            logging.warning(
                "File locking/unlocking using fcntl not avaliable on %s.",
                sys.platform,
            )
        else:
            shared_lock = fcntl.LOCK_SH
            unlock_operation = fcntl.LOCK_UN
        #
        self.lock_operation = shared_lock
        self.unlock_operation = unlock_operation
        self.handler_class: Type[StreamHandler] = FileReader
        self.open_mode = "r"
        if modify_in_place and replace_value_mode:
            self.lock_operation = exclusive_lock
            self.handler_class = FileWriter
            self.open_mode = "r+"
        #

    def open(self, encoding: str = "utf-8") -> IO[Any]:
        """Wrapper around open()"""
        return open(self.file_name, mode=self.open_mode, encoding=encoding)

    @staticmethod
    def execute_lock_op(file_handle: IO, operation: int) -> None:
        """Execute the supplied lock operation"""
        try:
            fcntl.flock(file_handle, operation)
        except NameError:
            pass
        #

    def lock(self, file_handle: IO) -> None:
        """Lock the file"""
        self.execute_lock_op(file_handle, self.lock_operation)

    def unlock(self, file_handle: IO) -> None:
        """Lock the file"""
        self.execute_lock_op(file_handle, self.unlock_operation)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

dryjq.queries

Query parser and data structure address objects

Copyright (C) 2022 Rainer Schwarzbach

This file is part of dryjq.

dryjq is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

dryjq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import collections
import logging

from typing import Any, Dict, Iterator, List, Optional, Tuple

import yaml

from dryjq import access


# Token types

TOKEN_START = "QUERY START"
TOKEN_END = "QUERY END"
TOKEN_LITERAL = "LITERAL"
TOKEN_SEPARATOR = "SEPARATOR"
TOKEN_SUBSCRIPT_OPEN = "SUBSCRIPT OPENING"
TOKEN_SUBSCRIPT_CLOSE = "SUBSCRIPT CLOSING"
TOKEN_SPACING = "SPACING"
TOKEN_ASSIGNMENT = "ASSIGNMENT"
TOKEN_INTERNAL_JOINER = "INTERNAL JOINER"
TOKEN_NONE = "NONE"

KNOWN_TOKENS = (
    TOKEN_START,
    TOKEN_END,
    TOKEN_LITERAL,
    TOKEN_SEPARATOR,
    TOKEN_SUBSCRIPT_OPEN,
    TOKEN_SUBSCRIPT_CLOSE,
    TOKEN_SPACING,
    TOKEN_ASSIGNMENT,
    TOKEN_INTERNAL_JOINER,
    TOKEN_NONE,
)


class Token:

    """Represents a single token"""

    def __init__(self, kind: str, content: Optional[Any] = None) -> None:
        """Store the token kind and content"""
        if kind not in KNOWN_TOKENS:
            raise ValueError(f"Unknown token kind {kind!r}")
        #
        self.__kind = kind
        self.__content = content

    @property
    def kind(self):
        """Return the kind"""
        return self.__kind

    @property
    def content(self):
        """return the content"""
        if self.__content is None:
            return f"<{self.kind}> token"
        #
        return self.__content


class Tokenizer:

    """Tokenize a query string for the parser"""

    quotes = (34, 39)  # Double and single quotes

    characters_lookup: Dict[str, Tuple[int, ...]] = {
        TOKEN_ASSIGNMENT: (61,),  # Equals sign
        TOKEN_SEPARATOR: (46,),  # Dot
        TOKEN_SPACING: (9, 10, 13, 32),  # Tab, LF, CR, Space
        TOKEN_SUBSCRIPT_OPEN: (91,),  # Opening square bracket
        TOKEN_SUBSCRIPT_CLOSE: (93,),  # Closing square bracket
    }

    def __init__(self) -> None:
        """Initialize the tokenizer internal state"""
        self.__open_quote: Optional[int] = None
        self.__in_subscript = False
        self.__current_literal: List[int] = []
        self.__last_token: Token = Token(TOKEN_NONE)
        # Build token lookup map
        self.__token_lookup = {}
        for (token_kind, characters) in self.characters_lookup.items():
            for single_character in characters:
                self.__token_lookup[single_character] = token_kind
            #
        #

    def add_token(self, kind: str, content: Optional[str] = None) -> Token:
        """Create a token, set self.__last_token to it
        and return it
        """
        self.__last_token = Token(kind, content=content)
        return self.__last_token

    def add_current_literal(self) -> Token:
        """Return a new token from the current literal
        if it exists, else an INTERNAL_JOINER token
        """
        if not self.__current_literal:
            return Token(TOKEN_INTERNAL_JOINER)
        #
        literal_data = "".join(
            chr(charcode) for charcode in self.__current_literal
        )
        logging.debug("Literal data: %r", literal_data)
        new_token = self.add_token(
            TOKEN_LITERAL,
            content=literal_data,
        )
        self.__current_literal.clear()
        return new_token

    def reset(self) -> None:
        """Reset the tokenizer"""
        self.__open_quote = None
        self.__in_subscript = False
        self.__current_literal.clear()
        self.__last_token = Token(TOKEN_NONE)

    def process_character(self, charcode: int) -> Iterator[Token]:
        """Yield a single token or more,
        according to current character state
        """
        token_kind = self.__token_lookup.get(charcode, TOKEN_LITERAL)
        if self.__open_quote:
            self.__current_literal.append(charcode)
            if charcode == self.__open_quote:
                yield self.add_current_literal()
                self.__open_quote = None
            #
        elif self.__in_subscript:
            if token_kind in TOKEN_SPACING:
                yield self.add_token(token_kind, chr(charcode))
                return
            #
            if token_kind == TOKEN_SUBSCRIPT_CLOSE:
                yield self.add_current_literal()
                yield self.add_token(token_kind, chr(charcode))
                self.__in_subscript = False
                return
            #
            if token_kind == TOKEN_SUBSCRIPT_OPEN:
                logging.warning(
                    "Possible error: found unexpected %s character %r"
                    " in subscript",
                    token_kind,
                    chr(charcode),
                )
                self.__current_literal.append(charcode)
            #
        elif token_kind == TOKEN_SUBSCRIPT_OPEN:
            yield self.add_current_literal()
            yield self.add_token(token_kind, chr(charcode))
            self.__in_subscript = True
            return
        #
        if token_kind in (TOKEN_ASSIGNMENT, TOKEN_SEPARATOR, TOKEN_SPACING):
            yield self.add_current_literal()
            yield self.add_token(token_kind, chr(charcode))
        elif charcode in self.quotes:
            if not self.__current_literal:
                self.__open_quote = charcode
            #
            self.__current_literal.append(charcode)
        else:
            self.__current_literal.append(charcode)
        #

    def tokenize(self, original_query: str) -> Iterator[Token]:
        """Yield tokens from a given query string"""
        self.reset()
        query_characters = collections.deque(original_query)
        yield self.add_token(TOKEN_START)
        while query_characters:
            for current_token in self.process_character(
                ord(query_characters.popleft())
            ):
                if current_token.kind != TOKEN_INTERNAL_JOINER:
                    logging.debug(
                        "%-20s -> %r",
                        current_token.kind,
                        current_token.content,
                    )
                    yield current_token
                #
            #
            if self.__last_token.kind == TOKEN_ASSIGNMENT:
                break
            #
        #
        last_literal_token = self.add_current_literal()
        if last_literal_token.kind != TOKEN_INTERNAL_JOINER:
            yield last_literal_token
        #
        if self.__last_token.kind == TOKEN_ASSIGNMENT:
            # Append the remainder als a single literal
            yield Token(TOKEN_LITERAL, content="".join(query_characters))
        #
        if self.__open_quote:
            raise ValueError(f"Unclosed quote ({self.__open_quote})!")
        #
        if self.__in_subscript:
            raise ValueError("Unclosed subscript ([)!")
        #
        yield Token(TOKEN_END)


class Parser:

    """New Query parser"""

    append_component_before: Tuple[str, ...] = (
        TOKEN_ASSIGNMENT,
        TOKEN_END,
        TOKEN_SEPARATOR,
        TOKEN_SUBSCRIPT_CLOSE,
        TOKEN_SUBSCRIPT_OPEN,
    )

    def __init__(self) -> None:
        """Initialize the tokenizer internal state"""
        self.__started = False
        self.__in_subscript = False
        self.__cache_remaining = False
        self.__collected_literals: List[str] = []
        self.__collected_components: List[access.PathComponent] = []

    def reset(self) -> None:
        """Reset the parser"""
        self.__started = False
        self.__in_subscript = False
        self.__cache_remaining = False
        self.__collected_literals.clear()
        self.__collected_components.clear()

    def __append_path_component(self, must_exist: bool = False) -> None:
        """Add a path component from the collected literals"""
        if not self.__collected_literals:
            if must_exist:
                raise ValueError("Empty path component")
            #
            return
        #
        # Join adjacent literals, but yaml load them
        # separately to enable mixing different quotes.
        # Allow other literals than strings.(only)
        # if exactly one literal was collected.
        if len(self.__collected_literals) > 1:
            component_index = "".join(
                yaml.safe_load(item) for item in self.__collected_literals
            )
        else:
            component_index = yaml.safe_load(self.__collected_literals[0])
        #
        self.__collected_components.append(
            access.PathComponent(
                component_index, in_subscript=self.__in_subscript
            )
        )
        self.__collected_literals.clear()

    def __consume_token_continue(self, token: Token) -> bool:
        """Consume one token and return False
        if this was the final iteration
        (i.e. an END token was encountered).
        """
        if not self.__started:
            # Allow spacing before the initial separator
            if token.kind in (TOKEN_START, TOKEN_SPACING):
                return True
            #
            if token.kind == TOKEN_SEPARATOR:
                self.__started = True
                return True
            #
            raise ValueError(
                f"The query must start with a {TOKEN_SEPARATOR} token!"
            )
        #
        if token.kind == TOKEN_LITERAL:
            self.__collected_literals.append(token.content)
        elif token.kind in self.append_component_before:
            self.__append_path_component()
            if token.kind == TOKEN_SUBSCRIPT_OPEN:
                self.__in_subscript = True
            elif token.kind == TOKEN_SUBSCRIPT_CLOSE:
                self.__in_subscript = False
            elif token.kind == TOKEN_ASSIGNMENT:
                self.__cache_remaining = True
            elif token.kind == TOKEN_END:
                return False
            #
        #
        return True

    def parse_query(
        self, original_query: str
    ) -> Tuple[access.Path, Optional[str]]:
        """Parse the given query and return a tuple
        containing a data structure address
        and - optionally - a new value
        """
        self.reset()
        replacement = None
        cached_tokens: List[Token] = []
        tokenizer = Tokenizer()
        logging.debug("Found tokens in query:")
        for token in tokenizer.tokenize(original_query):
            logging.debug("%-20s -> %r", token.kind, token.content)
            if self.__cache_remaining:
                cached_tokens.append(token)
                continue
            #
            if not self.__consume_token_continue(token):
                break
            #
        #
        logging.debug(cached_tokens)
        try:
            last_token = cached_tokens[0]
            if last_token.kind == TOKEN_END:
                logging.warning("Assuming empty assignment value")
                replacement = ""
            elif last_token.kind == TOKEN_LITERAL:
                replacement = last_token.content
                assert cached_tokens[1].kind == TOKEN_END
            else:
                raise ValueError(
                    f"Unexpected {last_token.kind} token"
                    f" {last_token.content!r}!"
                )
            #
        except IndexError:
            pass
        #
        logging.debug("Path components: %r", self.__collected_components)
        return (
            access.Path(*self.__collected_components),
            replacement,
        )


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:
